library grizzly.series.array.common;

import 'dart:math' as math;
import 'dart:collection';
import 'package:grizzly_series/grizzly_series.dart';
import 'package:grizzly_primitives/grizzly_primitives.dart';
import '../array.dart';
import '../sample.dart';
import 'package:text_table/text_table.dart';

part 'stats.dart';

abstract class ArrayMixin<E> implements Array<E> {
  void removeMany(Iterable<E> values) {
    final set = new Set<E>.from(values);
    for (int i = length - 1; i >= 0; i--) {
      if (set.contains(this[i])) removeAt(i);
    }
  }

  set assign(Iterable<E> other) {
    if (length == other.length) {
      for (int i = 0; i < length; i++) this[i] = other.elementAt(i);
      return;
    }

    if (length > other.length) {
      removeRange(other.length);
      for (int i = 0; i < length; i++) this[i] = other.elementAt(i);
      return;
    }

    for (int i = 0; i < length; i++) this[i] = other.elementAt(i);
    addAll(other.skip(length));
  }
}

abstract class ArrayFixMixin<E> implements ArrayFix<E> {
  /// Sets all elements in the array to given value [v]
  set set(E v) {
    for (int i = 0; i < length; i++) {
      this[i] = v;
    }
  }

  set assign(Iterable<E> other) {
    if (other.length != length)
      throw new ArgumentError.value(other, 'other', 'Size mismatch!');

    for (int i = 0; i < length; i++) this[i] = other.elementAt(i);
  }
}

abstract class ArrayViewMixin<E> implements ArrayView<E> {
  Index1D get shape => new Index1D(length);

  IntPair<E> pairAt(int index) => intPair<E>(index, this[index]);

  Iterable<IntPair<E>> enumerate() =>
      Ranger.indices(length).map((i) => intPair<E>(i, this[i]));

  Iterable<int> get i => Ranger.indices(length);

  int count(E v) {
    int ret = 0;
    for (E item in this) {
      if (v != item) ret++;
    }
    return ret;
  }

  Array<E> unique() {
    final ret = new LinkedHashSet<E>();
    for (E v in this) {
      if (!ret.contains(v)) ret.add(v);
    }
    return makeArray(ret);
  }

  Int1D uniqueIndices() {
    final ret = new LinkedHashMap<E, int>();
    for (int i = 0; i < length; i++) {
      E v = this[i];
      if (!ret.containsKey(v)) {
        ret[v] = i;
      }
    }
    return new Int1D(ret.values);
  }

  /// Returns a new  [Int1D] containing first [count] elements of this array
  ///
  /// If the length of the array is shorter than [count], all elements are
  /// returned
  Array<E> head([int count = 10]) {
    if (length <= count) return makeArray(this);
    return slice(0, count);
  }

  /// Returns a new  [Int1D] containing last [count] elements of this array
  ///
  /// If the length of the array is shorter than [count], all elements are
  /// returned
  Array<E> tail([int count = 10]) {
    if (length <= count) return makeArray(this);
    return slice(length - count);
  }

  /// Returns a new  [Array] containing random [count] elements of this array
  ///
  /// If the length of the array is shorter than [count], all elements are
  /// returned
  Array<E> sample([int count = 10]) => makeArray(sampler<E>(this, count));

  @override
  StringArray toStringArray() => new String1D(this.map((e) => e.toString()));

  @override
  IntSeries<E> valueCounts(
      {bool sortByValue: false, bool descending: false, name: ''}) {
    final groups = new Map<E, int>();
    for (int i = 0; i < length; i++) {
      final E v = this[i];
      if (!groups.containsKey(v)) groups[v] = 0;
      groups[v]++;
    }
    final ret = new IntSeries<E>.fromMap(groups, name: name);
    // Sort
    if (sortByValue) {
      ret.sortByLabel(descending: descending);
    } else {
      ret.sortByValue(descending: descending);
    }
    return ret;
  }

  @override
  BoolArray ne(other) {
    final ret = new Bool1D.sized(length);
    if (other is E) {
      for (int i = 0; i < length; i++) {
        ret[i] = this[i] != other;
      }
    } else if (other is Iterable<E>) {
      if (other.length != length)
        throw new LengthMismatch(
            expected: length, found: other.length, subject: 'other');
      for (int i = 0; i < length; i++) {
        ret[i] = this[i] != other.elementAt(i);
      }
    } else {
      throw new UnsupportedError('Type not supported!');
    }
    return ret;
  }

  @override
  BoolArray eq(other) {
    final ret = new Bool1D.sized(length);
    if (other is E) {
      for (int i = 0; i < length; i++) {
        ret[i] = this[i] == other;
      }
    } else if (other is Iterable<E>) {
      if (other.length != length)
        throw new LengthMismatch(
            expected: length, found: other.length, subject: 'other');
      for (int i = 0; i < length; i++) {
        ret[i] = this[i] == other.elementAt(i);
      }
    } else {
      throw new UnsupportedError('Type not supported!');
    }
    return ret;
  }

  @override
  BoolArray operator >=(other) {
    final ret = new Bool1D.sized(length);
    if (other is E) {
      for (int i = 0; i < length; i++) {
        ret[i] = compareValue(this[i], other) >= 0;
      }
    } else if (other is Iterable<E>) {
      if (other.length != length)
        throw new LengthMismatch(
            expected: length, found: other.length, subject: 'other');
      for (int i = 0; i < length; i++) {
        ret[i] = compareValue(this[i], other.elementAt(i)) >= 0;
      }
    } else {
      throw new UnsupportedError('Type not supported!');
    }
    return ret;
  }

  @override
  BoolArray operator >(other) {
    final ret = new Bool1D.sized(length);
    if (other is E) {
      for (int i = 0; i < length; i++) {
        ret[i] = compareValue(this[i], other) > 0;
      }
    } else if (other is Iterable<E>) {
      if (other.length != length)
        throw new LengthMismatch(
            expected: length, found: other.length, subject: 'other');
      for (int i = 0; i < length; i++) {
        ret[i] = compareValue(this[i], other.elementAt(i)) > 0;
      }
    } else {
      throw new UnsupportedError('Type not supported!');
    }
    return ret;
  }

  @override
  BoolArray operator <(other) {
    final ret = new Bool1D.sized(length);
    if (other is E) {
      for (int i = 0; i < length; i++) {
        ret[i] = compareValue(this[i], other) < 0;
      }
    } else if (other is Iterable<E>) {
      if (other.length != length)
        throw new LengthMismatch(
            expected: length, found: other.length, subject: 'other');
      for (int i = 0; i < length; i++) {
        ret[i] = compareValue(this[i], other.elementAt(i)) < 0;
      }
    } else {
      throw new UnsupportedError('Type not supported!');
    }
    return ret;
  }

  @override
  BoolArray operator <=(other) {
    final ret = new Bool1D.sized(length);
    if (other is E) {
      for (int i = 0; i < length; i++) {
        ret[i] = compareValue(this[i], other) <= 0;
      }
    } else if (other is Iterable<E>) {
      if (other.length != length)
        throw new LengthMismatch(
            expected: length, found: other.length, subject: 'other');
      for (int i = 0; i < length; i++) {
        ret[i] = compareValue(this[i], other.elementAt(i)) <= 0;
      }
    } else {
      throw new UnsupportedError('Type not supported!');
    }
    return ret;
  }

  bool operator ==(other) {
    if (other is Iterable<E>) {
      if (other.length != length) return false;
      for (int i = 0; i < length; i++) {
        if (this[i] != other.elementAt(i)) return false;
      }
      return true;
    }
    return false;
  }

  String toString() {
    final tab = table(Ranger.indices(length).map((i) => i.toString()).toList());
    tab.row(toStringArray().toList());
    return tab.toString();
  }
}
